##############################################################################
# DBMap.py - Class/Database mapping information
#    classes:
#	FieldMap:           database field/ class member variable mapping
#       DBMap:		    database table, id value, class, and fieldmap
#
#  (c) 7x Group, LLC.
#            author:  Chris Little, Larry Tuck, and Raymond Billings
#            created: 7/19/2001
#
##############################################################################
import types		# required for type comparisons


##############################################################################
# FieldMap
#
# Responsibilies:
#    - Map object attributes to table fields for a single class.
#
# Maps attributes to database fields
#
# dictionary:  Key            = property name
#              Value tuple[0] = column name
#                    tuple[1] = type: int, string, date, float, binary
##############################################################################
class FieldMap:
    '''Implements support for tying object properties to a DB'''
    
    INTEGER='int'
    FLOAT='float'
    STRING='string'
    DATE='date'
    BINARY='binary'

    def __init__(self, fieldMap = {}):
        self.fieldMap = fieldMap

    def listColumnsAsString(self):
        '''Get a comma seperated list of fields'''

        s = ""
        separator = ""
        for property in self.fieldMap.keys():
            s = s + separator + self.fieldMap[property][0]
            separator = ","

        return s

    def listColumns(self):
	"""get database columns as a list"""
	fieldList=[]
	for i in self.fieldMap.values():
	    fieldList.append(i[0]);
	return fieldList

    def listProperties(self):
	"""get properties as a list"""
	return self.fieldMap.keys()

    def getType(self,property):
	"""return type given property, return empty string if not found"""
	tp=''
	if self.fieldMap.has_key(property):
	    tp=self.fieldMap[property][1]
	return tp

    def getInitValue(self,property):
	"""return initialization value, given property"""
	tp=self.fieldMap[property][1]
	if tp==FieldMap.INTEGER:
	    return 0
	elif tp==FieldMap.FLOAT:
	    return 0.0
	else:
	    return ''

    def getColumn(self,property):
	"""return column name, given property"""
	if self.fieldMap.has_key(property):
	    return self.fieldMap[property][0]
	else:
	    return ''

    def getValue(self,key):
	"return the tuple at the given key"

	return self.fieldMap[key]

    def union(self,fm):
        "union the given fieldmap to the current fieldmap"

	keys=fm.listProperties()
	for i in keys:
	    self.fieldMap[i]=fm.getValue(i)

    def dump(self):
	print self.fieldMap

    def getBinaryColumn(self):
	'''return first binary column within the map'''
	for i in self.fieldMap.keys():
	    if self.fieldMap[i][1]==FieldMap.BINARY:
		return self.fieldMap[i][0]
	return ''

    def getBinaryProperty(self):
	'''return first binary property within the map'''
	for i in self.fieldMap.keys():
	    if self.fieldMap[i][1]==FieldMap.BINARY:
		return i
	return ''

#End class FieldMap
        

##############################################################################
#  DBMap
#
#  Responsibilities:
#     Maintain the mapping of object attributes to DB tables and fields for
#     an object and it's super classes.
#
#  Properties:
#
#    idLookup:  lookup string for the primary key
#    oidName: Name of object id, 'id' by default
#    keyName: Name of table primary key, 'ID' by default
#
#    dbMap:
#
#      List of tuples containing database information about class and class
#      hierarchy
#
#	tuple[0] = classname
#       tuple[1] = tablename
#       tuple[2] = Key Insert value string
#       tuple[3] = fieldmap for given table
#
#  Methods:
#     constructor:            initializes property
#     append:                 allow new table information to be appended
#     listTables:             return list of tables
#     getIdValueByTablename:  return id value for insert
#     getFieldMapByTablename: return appropriate field map
#     buildViewName:          build bottom level view name
#     buildViewFieldMap:      build entire fieldmap
#   
##############################################################################
class DBMap:
    "Map class information to database"

    def __init__(self):
        """constructor:  initialize map"""
	self.dbMap = []
	self.idLookup = ""
	self.oidName='id'
	self.keyName='ID'


    def append(self,classname,tablename="",idValue="",fieldmap=None):
	"""append given tuple to map"""
	self.dbMap.append((classname,tablename,idValue,fieldmap))


    def listTables(self):
	"""return list of tables within the db map"""
	rv = []
	for tuple in self.dbMap:
	    rv.append(tuple[1])

	return rv
	

    def listProperties(self):
	"""list all properties in dbmap"""
	fieldmap=self.buildViewFieldMap()
	return fieldmap.listProperties()

    def getIdValueByTablename(self,tablename):
	"""return the insert id value given table name"""
	for tuple in self.dbMap:
	    if tuple[1]==tablename:
		return tuple[2]

	return ""


    def getKeyName(self):
	"""return the primary key column name"""
	return self.keyName


    def getRootTable(self):
	"""return the root table of hierarchy"""
	return self.dbMap[0][1]


    def getBottomTable(self):
	"""return the bottom table of hierarchy"""
	l=len(self.dbMap)
	return self.dbMap[l-1][1]


    def setKeyName(self,name):
	"""set the primary key column name"""
	self.keyName=name


    def getOidName(self):
	"""return the object id name"""
	return self.oidName


    def setOidName(self,name):
	"""set the object id property name"""
	self.oidName=name


    def getIdLookup(self):
	"""return the lookup value for the primary key"""
	return self.idLookup

    def setIdLookupValue(self,idLookup):
	"""set the idLookup value"""
	self.idLookup=idLookup

    def getColumnName(self,property):
	"""return column name given property, '' if doesn't exist"""
	for tuple in self.dbMap:
	    fieldmap=tuple[3]
	    column=fieldmap.getColumn(property)
	    if column:
		return column
	return ''

    def getType(self,property):
	"""return type given property, '' if doesn't exist"""
	for tuple in self.dbMap:
	    fieldmap=tuple[3]
	    tp=fieldmap.getType(property)
	    if tp:
		return tp
	return ''

    def getFieldMapByTablename(self,tablename):
	"""return the field map by the given tablename"""

	for tuple in self.dbMap:
	    if (tuple[1]==tablename):
		return tuple[3]

	return {}

    def getClassNameByTablename(self,tablename):
	"""return the classname by the given tablename"""

	for tuple in self.dbMap:
	    if (tuple[1]==tablename):
		return tuple[0]
	return ''

    def buildViewName(self):
	"""construct a view name: it is the last table in list preceeded by v_"""
	
	lastTable=""
	for tuple in self.dbMap:
	    lastTable=tuple[1]

	return "v_"+lastTable

    def buildViewFieldMap(self):
	"""construct view fieldmap: all attributes of all tables"""
	
	fieldMap=FieldMap({})
	for tuple in self.dbMap:
	    fm=tuple[3]
	    fieldMap.union(fm)

	return fieldMap

    def dump(self):
	"""output map to screen"""
	print "idLookup: " + self.getIdLookup()
	print "DBMap:"
	for tuple in self.dbMap:
	    print 'CLASSNAME: ' + tuple[0]
	    print 'TABLENAME: ' + tuple[1]
            print 'KEY VALUE: ' + tuple[2]
	    tuple[3].dump()
	print ""

#END Class DBMap


