##############################################################################
# Persist.py - Persistence module.
#    classes:
#       PersistenceObject:  database functionality, parent class for 
#                           all persistence objects
#
#  (c) 7x Group, LLC.
#            author:  Chris Little, Larry Tuck, and Raymond Billings
#            created: 7/19/2001
#
##############################################################################
import types		# required for type comparisons
import string		# for upper function
from DBMap import *
from Query import *
from DBAccess import *


##############################################################################
#  PersistentObject - The base class for classes that implement persistence .
#
#  Resposibilities:
#      - Abstract root class for all peristent objects.
#        This class DEFINES the interface for ALL persistent objects.
#        This class implements methods to insert, restore, update, and delete
#            objects from the persistent store.
#      
##############################################################################
class PersistentObject:
    '''Base class for implementing object persistence.'''


    ######################################################################
    # Constructor
    ######################################################################
    def __init__(self,appName=""):
        self.dbMap = DBMap()
	self.__appName=appName
	self.helper=None
	self.db=None
	self.setOidMethod=None
	self.getOidMethod=None
        

    ######################################################################
    # setAppName
    #
    # set the name of the application - connect string will also be
    # reset
    ######################################################################
    def setAppName(self,appName):
	"""set the application name"""
	self.__appName=appName


    ######################################################################
    # getAppName
    #
    ######################################################################
    def getAppName(self):
	"""return the name of the application"""
	return self.__appName


    ######################################################################
    # getDB
    #
    ######################################################################
    def getDB(self):
	"""return the access to db"""
	return self.db


    ######################################################################
    # setDB
    #
    ######################################################################
    def setDB(self,db):
	"""set the access to db"""
	self.db=db


    ######################################################################
    # setHelper
    #
    # Set the helper attribute, which references a Helper object
    ######################################################################
    def setHelper(self,helper):
	"""set the helper attribute with the given helper object"""
	self.helper=helper

    ######################################################################
    # getHelper
    #
    # return the associated helper object
    ######################################################################
    def getHelper(self):
	"""return the associated helper object"""
	return self.helper


    ######################################################################
    # getClassName
    #
    # return classname
    ######################################################################
    def getClassName(self):
	"""return classname"""
	return self.__class__.__name__


    ######################################################################
    # setReferencesToOidMethods
    #
    ######################################################################
    def setReferencesToOidMethods(self,getMethod,setMethod):
        self.setOidMethod=setMethod
	self.getOidMethod=getMethod


    ######################################################################
    # setObjectId(value)
    #
    ######################################################################
    def setObjectId(self,value):
	self.setOidMethod(value)


    ######################################################################
    # getObjectId()
    #
    ######################################################################
    def getObjectId(self):
	return self.getOidMethod()


    ######################################################################
    #  restoreFromTuple(self,tuple,fieldMap)
    #
    #  REQUIRES:
    #      The order of the values in the tuple MUST match the order of the
    #      values in the properties in the fieldMap.  (The DCOracle connection
    #      does not provide a method for getting the value by column name.)
    #
    #  Given a field map and a tuple, restore properties in map
    ######################################################################
    def restoreFromTuple(self, tuple, fieldMap=None):
	'''Restore object from tuple and field map'''

        # if fieldMap is not specified, it means we are restoring from the
        # view.  Build the views fieldmap.
	if (fieldMap == None):
	    fieldMap = self.dbMap.buildViewFieldMap()

        properties=fieldMap.listProperties()
        index=0
        for property in properties:
	    value=tuple[index]

            # Replace Oracle 'missing values'
	    #if (value==Missing.Value):
	    if value is None:
	 	value=fieldMap.getInitValue(property)

            # Convert integer values.
	    if (fieldMap.getType(property)=="int"):
	        if (type(value)!=types.IntType and value!=""):
		    value=int(value)

            # Convert float values
	    elif (fieldMap.getType(property)=="float"):
	        if (type(value)!=types.FloatType and value!=""):
		    value=float(value)

            # Now set the property on ourself with the value.
            setattr(self,property,value)

            # Move to the next index in the tuple.
            index=index + 1

	
    ######################################################################
    #  restoreFromResult(self,rs,fieldMap)
    #
    #  Given a field map and a result, restore properties in map
    #  rs can be either a single tuple, or a list of tuples or lists
    #
    #  if a list, only the first element will be restored
    #
    #  A field map would only get passed in from within this class
    #  and would be used to restore a subset of the attributes from
    #  the result set.  Look at the restore() method for further
    #  detail.
    ######################################################################
    def restoreFromResult(self, rs, fieldMap=None):
        '''Restore the object from tuple or top row of a resultSet.'''

        # if fieldMap is not specified, it means we are restoring from the
        # view.  Build the views fieldmap.
	if (fieldMap == None):
	    fieldMap = self.dbMap.buildViewFieldMap()

	# set properties off of first of list of dictionaries, tuples, or lists
	if (type(rs) == types.ListType and rs):
            self.restoreFromTuple(rs[0], fieldMap)

	# set properties given no result
	else:
            for property in fieldMap.listProperties():
	 	initValue=fieldMap.getInitValue(property)
		setattr(self,property,initValue)


    ######################################################################
    #  PersistentObject::restore
    #
    #  restore the object from the persistent store.
    ######################################################################
    def restore(self):
        '''Restore the object from the persistent store.'''

	#  Connection
	conn=self.getConnection()

	tables=self.listTables()
	for t in tables:
	    query=RestoreQuery(obj=self,access=self.db,tablename=t)
	    query.execute(conn)
	conn.close()


    ######################################################################
    #  PersistentObject::getId
    #
    #  abstract method, getId, returns 0, needs to be overridden in 
    #  subclass
    ######################################################################
    def getId(self):
	return 0


    ######################################################################
    #  PersistentObject::save
    #
    #  save the object (insert or update)
    ######################################################################
    def save(self):
	'''save the object (insert or update)'''
	if self.getId():
	    self.update()
	else:
	    self.insert()


    ######################################################################
    #  PersistentObject::insert
    #
    #  Insert the object into the persistent store.  
    #  Set the id of this object
    ######################################################################
    def insert(self):
        '''Insert the object into the persistent store.'''

        #  Connection
	conn=self.getConnection()

	tables=self.listTables()
	for t in tables:
	    query=InsertQuery(obj=self,tablename=t,access=self.db)
	    query.execute(connection=conn)

	# set the id of the object
	query=IdLookupQuery(access=self.db,obj=self)
	oid=query.asValue(connection=conn)
	if oid:
	    setattr(self,self.getOidName(),int(oid))
	conn.close()
        

    ######################################################################
    #  PersistentObject::update
    #
    #  Update the object in the persistent store.
    ######################################################################
    def update(self):
        '''Update the object in the persistent store.'''

        #  Connection
	conn=self.getConnection()

	tables=self.listTables()
	for t in tables:
	    query=UpdateQuery(access=self.db,
		obj=self,tablename=t)
	    query.execute(conn)
	conn.close()


    ########################################################################
    #  PersistentObject::delete()
    #
    #  Delete the object from the persistent store.
    ########################################################################
    def delete(self):
        '''Delete the object from the from the persistent store.'''

        #  Connection
	conn=self.getConnection()

	tables=self.dbMap.listTables()
	for t in tables:
	    query=DeleteQuery(access=self.db,obj=self,tablename=t)
	    query.execute(conn)

	self.deleteAssociations(conn)
	conn.close()


    ########################################################################
    #  deleteAssociations(conn)
    #
    #  Abstract method to delete associations.  The correct delete
    #  queries will have to be put together and executed in the
    #  deleteAssociations method written within the subclass
    ########################################################################
    def deleteAssociations(self,connection):
	"""abstract method for deletion of associated rows"""
	return


    ########################################################################
    # getConnection()
    #
    # (I created this method to consolidate the connection code
    #  which is driver speciifc.)
    #
    # Returns a connection.
    ######################################################################
    def getConnection(self):
	'''return oracle connection'''

	if not self.db:
	    self.db=DBAccess(self.getAppName())
        return self.db.getConnection()


    ######################################################################
    #  getViewName
    #
    #  return the view name , delegates to dbMap
    ######################################################################
    def getViewName(self):
	"""return view name for this object"""
	return self.dbMap.buildViewName()


    ######################################################################
    #  getViewFieldMap
    #
    #  return the fieldmap for the view
    ######################################################################
    def getViewFieldMap(self):
	"""return view fieldmap for this object"""
	return self.dbMap.buildViewFieldMap()


    ######################################################################
    #  getColumnName
    #
    #  return the column name for the given property name
    ######################################################################
    def getColumnName(self,propertyName):
	"""return column name for the given propertyName"""
	return self.dbMap.getColumnName(propertyName)


    ######################################################################
    #  getType
    #
    #  return the data type given the property name
    ######################################################################
    def getType(self,propertyName):
	"""return type for the given propertyName"""
	return self.dbMap.getType(propertyName)


    ######################################################################
    #  listTables
    #
    #  return list of tables that make up this object
    ######################################################################
    def listTables(self):
	"""return list of tables that make up this object"""
	return self.dbMap.listTables()

    ######################################################################
    #  getKeyColumn
    #
    #  return key column name for this object
    ######################################################################
    def getKeyColumn(self):
	"""return key column name for this object"""
	return self.dbMap.getKeyName()

    ######################################################################
    #  getOidName
    #
    #  return key column name for this object
    ######################################################################
    def getOidName(self):
	"""return object id property name for this object"""
	return self.dbMap.getOidName()

    ######################################################################
    #  getFieldMapByTablename
    #
    ######################################################################
    def getFieldMapByTablename(self,tablename):
	"""return classname for the given table"""
	return self.dbMap.getFieldMapByTablename(tablename)

    ######################################################################
    #  getClassNameByTablename
    #
    ######################################################################
    def getClassNameByTablename(self,tablename):
	"""return classname for the given table"""
	return self.dbMap.getClassNameByTablename(tablename)

    ######################################################################
    #  getIdValueByTablename
    #
    ######################################################################
    def getIdValueByTablename(self,tablename):
	"""return id value for the given table"""
	return self.dbMap.getIdValueByTablename(tablename)

    ######################################################################
    #  getIdLookup
    #
    ######################################################################
    def getIdLookup(self):
	return self.dbMap.getIdLookup()

    ######################################################################
    #  getRootTable
    #
    ######################################################################
    def getRootTable(self):
	return self.dbMap.getRootTable()

    ######################################################################
    #  getTableName
    #
    ######################################################################
    def getTableName(self):
	return self.dbMap.getBottomTable()

    ######################################################################
    #  get
    #
    ######################################################################
    def get(self,propertyname,default):
	"""general purpose get, if property name invalid then return default"""
	if hasattr(self,propertyname):
	    return getattr(self,propertyname)
	else:
	    return default

    ######################################################################
    #  getPropertiesAsDictionary
    #
    #  returns only business properties as defined in dbmap
    ######################################################################
    def getPropertiesAsDictionary(self):
	"""return all business properties as dictionary.  Only those
           properties within dbmap will be returned"""
	properties=self.dbMap.listProperties()
	rv={}
	for i in properties:
	    rv[i]=getattr(self,i)
	return rv


    ######################################################################
    #  getAsDictionary
    #
    #  returns all properties as dictionary
    ######################################################################
    def getAsDictionary(self):
	"""return all properties as dictionary."""
	return self.__dict__

#END class PersistentObject


