#############################################################################
#  Query.py
#
#  (c) 7x Group, LLC.
#            author:  Raymond Billings
#            created: 8/28/2001
#
#  Query - module to carry classes for construction and execution of
#          common retrieval queries
#        - the primary goal of this module is the removal of SQL 
#          generation and knowledge from other classes, it should
#          all be contained here except under very rare circumstances
#        - these classes interact very closely with the PersistentObject
#          class but have been separated out into separate module in order
#          to ensure separation of SQL
#
#  Classes:
#        QueryObject   - class to encapsulate object being queried on
#
#        Query         - superclass for query classes 
#                        (currently only has  one subclass)
#        RetrieveQuery - query for retrieval of one or more tuples or 
#                        objects
#        DeleteMultipleQuery
#                      - query to delete multiple rows
#        RestoreQuery  - query to restore a table worth of attributes
#                        to an object,  multiple RestoreQueries may
#                        be required to restore an entire object
#        SingleTableQuery
#        RestoreQuery  - for restore from table for a given id
#        InsertQuery   - for insert to table providing and id
#        UpdateQuery   - for update of table for a given id
#        DeleteQuery   - for delete from table for a given id
#        OverrideQuery - allow ability to override any standard query
#        SubQuery      - sub-query for exists and in
#        IdLookupQuery - query for looking up current id from database
#
#        Condition     - superclass for query criteria conditions
#        Criteria      - a condition consisting of multiple and/or clauses
#                        queries have a criteria property
#        Comparison    - a simple comparison operation
#        OverrideCondition - a condition which is basically a string
#                        override for maximum flexibility
#        ExistsCondition - condition for exists sub-query
#
#############################################################################
import types
import string
import exceptions
import DBAccess
from ObjectFactory import ObjectFactory


##############################################################################
#  QueryObject
#
#  Class to wrap object used for criteria and other query purposes
##############################################################################
class QueryObject:

    ######################################################################
    # constructor
    #
    #   obj    = reference object
    #   alias  = table alias for query purposes
    #   hidden = is this object hidden from FROM list or not 
    #            (an outer object would be hidden in sub queries)
    ######################################################################
    def __init__(self,obj,alias=''):
	self.obj=obj
	self.setAlias(alias)
	self.hidden=0


    ######################################################################
    #  __str__
    #
    #  return class name and alias
    ######################################################################
    def __str__(self):
	return "name: '%s' alias: '%s' hidden: %d" % \
		(self.obj.__class__.__name__,self.alias,self.hidden)


    ######################################################################
    #  setHidden
    ######################################################################
    def setHidden(self):
	self.hidden=1


    ######################################################################
    #  isHidden
    ######################################################################
    def isHidden(self):
	return self.hidden


    ######################################################################
    #  setAlias
    ######################################################################
    def setAlias(self,alias=''):
	if not alias:
	    self.alias=''
	elif string.find(alias,".")>=0:
	    self.alias=alias
	else:
	    self.alias=alias+"."


    ######################################################################
    #  getViewName
    #
    #  return the view name with alias if desired
    ######################################################################
    def getViewName(self,withAlias=0):
	"""return view name"""
	viewName=self.obj.getViewName()
	if withAlias:
	    viewName=self.alias+viewName
	return viewName


    ######################################################################
    #  getAlias
    ######################################################################
    def getAlias(self):
        return self.alias


    ######################################################################
    #  getViewFieldMap
    #
    #  return the fieldmap for the view representing this object
    ######################################################################
    def getViewFieldMap(self):
	"""return the fieldmap for the view represinting this object"""
	return self.obj.getViewFieldMap()


    ######################################################################
    # getColumnName
    #
    # return column name given a property name, add prefix if requested
    ######################################################################
    def getColumnName(self,propertyName,withAlias=0):
	"""return column name given property name"""
	columnName=self.obj.getColumnName(propertyName)

	if columnName and withAlias:
	    if self.alias:
	        columnName=self.alias+columnName
	    else:
	        viewname=self.obj.getViewName()
	        columnName=viewname+"."+columnName

	return columnName


    ######################################################################
    # Methods that delegate directly to the object
    ######################################################################
    def getType(self,propertyName):
	"""return column type"""
	columnType=self.obj.getType(propertyName)

	return columnType

    def getClassName(self):
	"""return name of class for object"""
	return self.obj.getClassName()

    def getIdLookup(self):
	return self.obj.getIdLookup()

    def getRootTable(self):
	return self.obj.getRootTable()

    def getKeyColumn(self):
	return self.obj.getKeyColumn()

    def getOidName(self):
	return self.obj.getOidName()

    def getFieldMapByTablename(self,tablename):
	return self.obj.getFieldMapByTablename(tablename)

    def getClassNameByTablename(self,tablename):
	return self.obj.getClassNameByTablename(tablename)

    def restoreFromResult(self,rs,fieldmap):
	self.obj.restoreFromResult(rs,fieldmap)

    def getIdValueByTablename(self,tablename):
	return self.obj.getIdValueByTablename(tablename)

    def get(self,property,default=None):
	return self.obj.get(property,default)

    def getTableName(self):
        return self.obj.getTableName()

    ######################################################################
    #  getColumns
    #
    #  build comma delimited string of columns with date conversion
    #
    #  If parameter is field map then use it, otherwise generate
    #  use entire view
    ######################################################################
    def getColumns(self,fieldmap=None,withAlias=0):
        '''Get a comma seperated list of columns with date conversion'''

	if not fieldmap:
	    fieldmap=self.getViewFieldMap()
	if withAlias:
	    colPrefix=self.getAlias()
	    if not colPrefix:
		colPrefix=self.obj.getViewName()+"."

        s = ""
        separator = ""
        for property in fieldmap.listProperties():
	    column = colPrefix + fieldmap.getColumn(property)
	    if fieldmap.getType(property) == "date":
                s = s + separator + Query.DATE_TO_STRING % column
	    else:
                s = s + separator + column
            separator = ","

        return s

# END QueryObject class




#----------------------------------------------------------------------------
#--  Query Classes
#----------------------------------------------------------------------------

#############################################################################
#  Query
#
#  Class to represent a general query.  Superclass for more specific 
#  query types
#############################################################################
class Query:

    # SQL Templates
    DATE_TO_STRING = "to_char(%s,'MM/DD/YYYY')"
    STRING_TO_DATE = "to_date('%s','MM/DD/YYYY')"
    UPPER = "upper(%s)"
    LIKE_CONTAINS = "%%%s%%"
    LIKE_STARTS_WITH="%%%s"


    ######################################################################
    # constructor
    ######################################################################
    def __init__(self,access,obj=None,criteria=None):

	if obj:
	    self.setObject(obj)
	self.criteria=criteria
	self.accessName=''
	self.db=None
	self.setAccess(access)


    ######################################################################
    # __str__
    ######################################################################
    def __str__(self):
	return self.getSQL()


    ######################################################################
    # getSQL
    ######################################################################
    def getSQL(self):
	return ''


    #########################################################################
    # Query::execute
    #
    # can be performed against an open connection or a new connection
    # will be instantiated
    #########################################################################
    def execute(self,connection=None):
	'''execute the query'''

	query=self.getSQL()
	return self.db.execute(query=query,connection=connection)


    ######################################################################
    # setters and getters
    ######################################################################
    def setCriteria(self,criteria):
	"""set criteria"""
	self.criteria=criteria

    def getCriteria(self):
	"""return the criteria for this query"""
	return self.criteria

    def setAccess(self,value):
	if type(value)==types.StringType:
	    self.setAccessName(value)
	else:
	    self.setDb(value)

    def setAccessName(self,value):
	if value:
	    self.db=DBAccess.DBAccess(value)
	else:
	    self.db=None
	self.accessName=value

    def getAccessName(self):
	return self.accessName

    def setDb(self,value):
	self.db=value
	if hasattr(value,'getLookupName'):
	     self.accessName=value.getLookupName()
	else:
	     raise QueryException, 'access must be valid DBAccess object'

    def getDb(self):
	return self.db


    ######################################################################
    # setObject
    #
    # Set query object
    ######################################################################
    def setObject(self,obj,alias=''):
	"""set query object"""
    	if obj.__class__.__name__=='QueryObject':
	    self.obj=obj
	else:
	    self.obj=QueryObject(obj)
	if alias:
	    self.obj.setAlias(alias)


    ######################################################################
    #  getColumns
    #
    #  delegate to QueryObject to return comma deliminated list of
    #  columns
    ######################################################################
    def getColumns(self):
        obj=self.obj
	return obj.getColumns(withAlias=1)
    

    ######################################################################
    #  __getFrom
    #
    #  return dictionary of views for from clause
    ######################################################################
    def __getFrom(self):
	fromSet={}
	k="%s %s" % (self.obj.getViewName(),self.obj.getAlias())
	fromSet[k]=self.obj.getAlias()
	
        if self.criteria:
	    self.criteria.getFrom(fromSet)

	return fromSet


    ######################################################################
    #  getFrom
    #
    #  return comma delimited list of view names to query from
    ######################################################################
    def getFromList(self):
	fromSet=self.__getFrom()
	fromList=''
	separator=''
	for i in fromSet.keys():
	    fromList=fromList + separator + i
	    separator=", "
	return fromList


    ######################################################################
    #  addCriteria
    #
    #  add criteria to query and return the added condition
    ######################################################################
    def addCriteria(self,left=None,right=None,op='=',negate=0,
		caseSensitive=0,condition=None):
	"""add criteria to query"""
	if not self.criteria:
	    self.criteria=Criteria(obj=self.obj,negate=0,conditions=[])

	rv=self.criteria.add(left=left,right=right,op=op,negate=negate,
		caseSensitive=caseSensitive,condition=condition)
	return rv


    ########################################################################
    #  isPostgres
    #
    #  Check if PostgreSQL database or not
    ########################################################################
    def isPostgres(self):
	'''check if PostgreSQL database or not'''
	dbtype=self.db.getDBType()
	if dbtype==DBAccess.DBAccess.POSTGRES:
	    return 1
	return 0


    ########################################################################
    #  isOracle
    #
    #  Check if Oracle database or not
    ########################################################################
    def isOracle(self):
	'''check if Oracle database or not'''
	dbtype=self.db.getDBType()
	if dbtype==DBAccess.DBAccess.ORACLE:
	    return 1
	return 0

#END class Query


#############################################################################
#  SingleTableQuery
#  
#  Class to represent a query against a single table.  It is possible
#  for an object to be stored in multiple tables, so instances of 
#  queries of this type may only hold the part of the object information
#
#############################################################################
class SingleTableQuery(Query):

    #########################################################################
    # constructor
    #########################################################################
    def __init__(self,access,obj,tablename='',idBasedCriteria=1):
	self.tablename=tablename
	self.idBasedCriteria=idBasedCriteria
	Query.__init__(self,access=access,obj=obj)
	if idBasedCriteria:
	    self.__setCriteria()


    ######################################################################
    # setObject
    #
    # Set query object - calls superclass set object and then sets
    # criteria based on object
    ######################################################################
    def setObject(self,obj,alias=''):
	"""set query object"""
	Query.setObject(self,obj,alias)
	if self.idBasedCriteria:
	    self.__setCriteria()
	self.__setFieldMap()
	self.__setBinaryColumn()


    ######################################################################
    # __setFieldMap
    #
    ######################################################################
    def __setFieldMap(self):
	if not self.tablename:
	    self.fieldmap=self.obj.getViewFieldMap()
	else:
	    self.fieldmap=self.obj.getFieldMapByTablename(self.tablename)


    ######################################################################
    # __setBinaryColumn
    #
    ######################################################################
    def __setBinaryColumn(self):
	if self.fieldmap.getBinaryProperty():
	    self.binaryProperty=self.fieldmap.getBinaryProperty()
	else:
	    self.binaryProperty=''


    #########################################################################
    # getTableName
    #
    # return the table or view name
    #########################################################################
    def getTableName(self):
	'''return the name of the table or view referenced by this query'''

	if self.tablename:
	    return self.tablename
	return self.obj.getViewName()


    #########################################################################
    # __setCriteria
    #
    # set criteria based on object id
    #########################################################################
    def __setCriteria(self):
	"""set the restore criteria"""
	self.obj.setAlias(self.getTableName())

	criteria=Criteria(obj=self.obj,negate=0,conditions=[])
	key=self.obj.getKeyColumn()
	idName=self.obj.getOidName()

	id=self.obj.get(idName,0)
	criteria.add(left=key,right=id,op=Condition.EQ)
	self.criteria=criteria


    ######################################################################
    #  convertTextForUpdate
    #
    #  The following conversions will be made:
    #      - '\n' will be converted to new line
    #      - single quote will be converted to 3 single quotes
    ######################################################################
    def convertTextForUpdate(self,str):
        s1 = string.replace(str,'\\n','\n')
        s2 = self.convertStringForUpdate(self,s1)
        return s2


    ######################################################################
    #  convertStringForUpdate
    #
    #  The following conversions will be made:
    #      - single quote will be converted to 3 single quotes
    ######################################################################
    def convertStringForUpdate(self,str):
        s1 = string.replace(str,"'","''")
        return s1


#End class SingleTableQuery



#############################################################################
#  RestoreQuery
#
#  Class to represent basic restore query to restore from a particular 
#  table (multiple restore queries may be required to put together an object)
#############################################################################
class RestoreQuery(SingleTableQuery):

    #########################################################################
    # constructor
    #########################################################################
    def __init__(self,access,obj,tablename=''):
	SingleTableQuery.__init__(self,access=access,
		obj=obj, tablename=tablename, idBasedCriteria=1)


    #########################################################################
    # getSQL
    #########################################################################
    def getSQL(self):
	"""return sql for this query"""

	s = "select %s from %s where %s\n" % \
	    (self.getColumns(), self.getTableName(), self.criteria.getSQL())

	return s
	

    #########################################################################
    #  getColumn
    #########################################################################
    def getColumns(self):
	"""return comma delimited list of columns"""

	s = ""
        separator = ""
	colPrefix=self.obj.getAlias()
        for property in self.fieldmap.listProperties():
            column = colPrefix + self.fieldmap.getColumn(property)
            if self.fieldmap.getType(property) == "date":
		s = s + separator + Query.DATE_TO_STRING % column
            else:
                s = s + separator + column
            separator = ","

        return s
	

    #########################################################################
    # execute
    #
    # can use an open connection, or will instantiate its own if it doesn't
    # get one
    #########################################################################
    def execute(self,connection=None):
	'''execute the query and restore the attributes of the object
	   from the query'''

	query=self.getSQL()
	rs=self.db.execute(query=query,connection=connection)
	self.obj.restoreFromResult(rs,self.fieldmap)

	return rs

#End class RestoreQuery


#############################################################################
#  InsertQuery
#
#  Class to represent basic insert query to part of an object (a tables worth)
#############################################################################
class InsertQuery(SingleTableQuery):

    #########################################################################
    # constructor
    #########################################################################
    def __init__(self,access,obj,tablename=''):
	SingleTableQuery.__init__(self,access=access,obj=obj,
		tablename=tablename,idBasedCriteria=0)
	self.__setIdValue()


    ######################################################################
    # __setIdValue
    #
    ######################################################################
    def __setIdValue(self):
	if not self.tablename:
	    self.idValue=0
	else:
	    self.idValue=self.obj.getIdValueByTablename(self.tablename)


    #########################################################################
    # getSQL
    #########################################################################
    def getSQL(self):
	"""return sql for this query"""

	s = "insert into %s (%s) values (%s)" % \
	  (self.getTableName(), self.getColumns(), self.getValues())

	return s
	

    #########################################################################
    #  getColumns
    #########################################################################
    def getColumns(self):
	"""return comma delimited list of columns"""
	return self.fieldmap.listColumnsAsString()
	

    #########################################################################
    #  getValues
    #########################################################################
    def getValues(self):
	'''return insertable list of values'''

        s = ""
        separator = ""
        oidName=self.obj.getOidName()
        for property in self.fieldmap.listProperties():
            if property==oidName and self.idValue:
                s = s + separator + self.idValue
            else:
                if self.fieldmap.getType(property) =="date":
                    s = s + separator + Query.STRING_TO_DATE % \
			self.obj.get(property,'')
                elif self.fieldmap.getType(property)=="string":
                    s = s + separator + "'%s'" % \
                        self.convertStringForUpdate(self.obj.get(property,''))
                elif self.fieldmap.getType(property)=="text":
                    s = s + separator + "'%s'" % \
                        self.convertTextForUpdate(self.obj.get(property,''))
		elif self.fieldmap.getType(property)=="binary":
		    s = s + separator + self.__getBinarySubstitute(property)
                else:
                    s = s + separator + `self.obj.get(property,0)`
            separator = ","
	return s


    #########################################################################
    # execute
    #
    # can be performed against an open connection or a new connection
    # will be instantiated
    #########################################################################
    def execute(self,connection=None):
        '''execute the query'''

        query=self.getSQL()
	if self.binaryProperty:
	    x=self.obj.get(self.binaryProperty)
	    rs = self.db.binExecute(query=query,connection=connection,x=x)
	else:
            rs = self.db.execute(query=query,connection=connection)
	return rs


    #########################################################################
    # __getBinarySubstitute
    #########################################################################
    def __getBinarySubstitute(self,property):
	if self.isPostgres():
	    rv='byteain(%s)' % (property)
	    rv=':x'
	else:
	    rv=':x'
	return rv


#End class InsertQuery


#############################################################################
#  UpdateQuery
#
#  Class to represent basic update query to part of an object (a tables worth)
#############################################################################
class UpdateQuery(SingleTableQuery):

    #########################################################################
    # constructor
    #########################################################################
    def __init__(self,access,obj,tablename=''):
	SingleTableQuery.__init__(self,access=access,obj=obj,
		tablename=tablename,idBasedCriteria=1)


    #########################################################################
    # getSQL
    #########################################################################
    def getSQL(self):
	"""return update sql"""

	s = "update %s set %s where %s" % \
	 (self.getTableName(), self.getAssignments(), 
		self.criteria.getSQL())

	return s
	

    #########################################################################
    #  getAssignments
    #########################################################################
    def getAssignments(self):
	'''return column=value assignments'''

        s = ""
        separator = ""
        for property in self.fieldmap.listProperties():
            dataType = self.fieldmap.getType(property)
            column = self.fieldmap.getColumn(property)
            if dataType == "date":
                s = s + separator + "%s = " % column + Query.STRING_TO_DATE % \
		    self.obj.get(property,'')
            elif dataType == "string":
		s = s + separator + "%s = '%s'" % \
		  (column,self.convertStringForUpdate(self.obj.get(property,'')))
            elif dataType == "text":
		s = s + separator + "%s = '%s'" % \
		  (column,self.convertTextForUpdate(self.obj.get(property,'')))
	    elif self.fieldmap.getType(property)=="binary":
		s = s + separator + "%s = :x" % \
		  (column)
            else:
		s = s + separator + "%s = %s" % (column,str(self.obj.get(property,0)))
            separator = ","

        return s

    #########################################################################
    # execute
    #
    # can be performed against an open connection or a new connection
    # will be instantiated
    #########################################################################
    def execute(self,connection=None):
        '''execute the query'''

        query=self.getSQL()
	if self.binaryProperty:
	    x=self.obj.get(self.binaryProperty)
	    rs = self.db.binExecute(query=query,connection=connection,x=x)
	else:
            rs = self.db.execute(query=query,connection=connection)
	return rs

#End class UpdateQuery


#############################################################################
#  DeleteQuery
#
#  Class to represent basic delete query to part of an object (a tables worth)
#############################################################################
class DeleteQuery(SingleTableQuery):

    #########################################################################
    # constructor
    #########################################################################
    def __init__(self,access,obj,tablename=''):
	SingleTableQuery.__init__(
	    self,access=access,obj=obj,tablename=tablename)


    #########################################################################
    # getSQL
    #########################################################################
    def getSQL(self):
	"""return delete sql"""

	s = "delete from %s where %s" % \
	     (self.getTableName(), self.criteria.getSQL())
	return s
	
#End class DeleteQuery


#############################################################################
#  RetrieveQuery
#
#  Class to represent a general retrieval query.  
#############################################################################
class RetrieveQuery(Query):
    
    #########################################################################
    # constructor
    #
    #  access     = name or DBAccess object to reference for database access
    #  obj        = object type to create with this query
    #  criteria   = criteria to use with this query
    #  selectColumns = columns to use for selection, default is all
    #  orderby    = columns to use for order by clause in query
    #  accessName (deprecated): name to reference for database access
    #########################################################################
    def __init__(self,access=None,obj=None,criteria=None,selectColumns=None,
	orderby=None,accessName=''):
	if not access:
	    access=accessName
	Query.__init__(self,access,obj,criteria)

	self.selectColumns=self.__makeList(selectColumns)
	self.orderby=self.__makeList(orderby)


    ########################################################################
    #  getSQL
    #
    #  Return sql for query, optional start and end row parameters
    #  allow for query by row number
    ########################################################################
    def getSQL(self,startRow=0,endRow=0):
	"""return sql for this query"""

	s='select '
	if not self.selectColumns:
	    if startRow:
		s=s+' * '
	    else:
	        s = s + self.getColumns() + "\n"        # select
	else:
	    separator=''
	    for i in self.selectColumns:
	        s=s+separator+self.obj.getColumnName(i,withAlias=1)
	        separator=', '


	if not startRow:
	    s = s + "from " + self.getFromList() + "\n"       # from
	elif endRow:
	    c1=Criteria(obj=self.obj,conditions=[])
	    if self.criteria:
	        c1.add(condition=self.criteria)
	    c1.add(left='rownum',right=endRow,op=Condition.LE)
	    q1=RetrieveQuery(obj=self.obj,criteria=c1)

	    c2=Criteria(obj=self.obj,conditions=[])
	    if self.criteria:
	        c2.add(condition=self.criteria)
	    c2.add(left='rownum',right=startRow,op=Condition.LT)
	    q2=RetrieveQuery(obj=self.obj,criteria=c2)
	    s = s + "from (%s) minus (%s)\n" % (q1.getSQL(), q2.getSQL())
	else:
	    raise QueryException, "start row specified but not end row"

	if not startRow:
	    if self.criteria:
	        conditionString = self.criteria.getSQL()
	        if conditionString:
	            s = s + " where "
	            s = s + conditionString + "\n"    # where
	    if endRow:
		if self.criteria:
		    s = s + " and "
		else:
	            s = s + " where "
		s = s + "rownum <= %s \n" % str(endRow)

	if self.orderby:
	    s = s + "order by "
	    separator=''
	    for i in self.orderby:
		if type(i)==types.IntType:
		    s=s+separator+str(i)
		else:
		    if startRow:
		        s=s+separator+self.obj.getColumnName(i,withAlias=0)
		    else:
		        s=s+separator+self.obj.getColumnName(i,withAlias=1)
		separator=", "

	return s


    ########################################################################
    #  filterByRownum
    #
    #  given a result set, and start/end row numbers, return a list
    #  of results that fall within the range
    ########################################################################
    def filterByRownum(self,rs,startRow=0,endRow=0):
	'''
            given a result set, and start/end row numbers, 
	    return a list of results that fall within the range
	'''
	if endRow<1:
	    return rs

	if startRow<1:
	    startRow=1
	startRow=startRow-1

	return rs[startRow:endRow]


    ########################################################################
    #  asTuples
    #
    #  Return list of tuples from resultset
    #  
    #  Notes: 
    #  
    #  1.  This is a very important method.  All of the other query
    #      retrieval methods (asValues, asObjects, asObject, asValue) call
    #      this method in order to obtain the initial results to work from.
    #
    #  2.  Oracle can handle selects by row numbers whereas postgres
    #      cannot, therefore the filtering by row number is done after
    #      the query for those systems that cannot let the database 
    #      handle it.
    ########################################################################
    def asTuples(self,connection=None,startRow=0,endRow=0):
	'''Return the results of the query as a list of tuples (or lists)
	   
	   Arguments:
		connection:  if a connection is provided then a 
		             new connection will not be opened and closed
		startRow:   
		endRow:	     if start and end rows are specified then only
			     the results that fall within the range will
			     be returned.  An endRow can be specified 
			     by itself, without a startRow, but not vice
			     versa.

	   Return Value:     list of tuples or lists depending on the
			     the underlying database system.
	'''
	   
	#  Filter by row number after query
	if self.isPostgres() and endRow>0:
	    query=self.getSQL(startRow=0,endRow=0)
	    tmpRs=self.db.execute(query=query,connection=connection)
	    rs=self.filterByRownum(rs=tmpRs,startRow=startRow,endRow=endRow)

	#  Filter by row number in query
	else:
	    query=self.getSQL(startRow,endRow)
	    rs=self.db.execute(query=query,connection=connection)

	return rs


    ########################################################################
    #  asObjects
    #
    #  Return list of objects from this query
    ########################################################################
    def asObjects(self,connection=None,startRow=0,endRow=0):
        '''Return list of objects from a resultset.'''

	factory = ObjectFactory(self.getAccessName(),db=self.db)
	rs=self.asTuples(connection=connection,startRow=startRow,endRow=endRow)

	appName=self.getAccessName()
        objectList=[]
	if rs:
          for row in rs:
            object = factory.instantiate(self.obj.getClassName())
	    if hasattr(object,'setAppName'):
	        object.setAppName(appName)
	    else:
		raise QueryException, "Unknown application name for " \
			+ obj.getClassName()
            object.restoreFromTuple(row)
	    helper=object.getHelper()
	    if helper:
		object=helper.restore(object)
		object.setAppName(appName)
            objectList.append(object)

        return objectList


    #########################################################################
    # asObject
    #########################################################################
    def asObject(self,connection=None):
	'''return first object returned, or None'''
	lst=self.asObjects(connection=connection,endRow=1)
	if lst:
	    return lst[0]

	return None


    #########################################################################
    # asValue
    #########################################################################
    def asValue(self,connection=None):
	'''return first column result from first row returned, or None'''
	lst=self.asTuples(connection=connection,endRow=1)
	if lst:
	    return lst[0][0]
	return None


    #########################################################################
    # hasResult
    #########################################################################
    def hasResult(self,connection=None):
	'''return true or false depending on if results are returned or not'''
	lst=self.asTuples(connection=connection,endRow=1)
	if lst:
	    return 1
	return 0


    #########################################################################
    # asValues
    #########################################################################
    def asValues(self,connection=None):
	'''return first column result from first row returned, or None'''
	lst=self.asTuples(connection=connection)
	valueLst=[]
	for i in lst:
	    valueLst.append(i[0])
	return valueLst


    #########################################################################
    # addSearch
    #
    #########################################################################
    def addSearch(self,exact=0,**parms):
	"""set up a basic search condition for this query,
	   this method allows for easy set up of query for
           common search ability"""
	criteria=Criteria(obj=self.obj,conditions=[])
	criteria.addSearchFromDict(exact=exact,parms=parms)
	if criteria.conditions:
	    self.criteria=criteria


    #########################################################################
    # setOrderBy
    #
    #########################################################################
    def setOrderBy(self,orderby):
	'''Set order by columns of query'''
	self.orderby=self.__makeList(orderby)


    #########################################################################
    # setOrderBy
    #
    #########################################################################
    def setSelectColumns(self,selectColumns):
	'''Set select columns for query'''
	self.selectColumns=self.__makeList(selectColumns)


    #########################################################################
    # __makeList
    #
    # Private method to make sure that whatever is provided gets 
    # returned as a list
    #########################################################################
    def __makeList(self,i):
	if not i:
	    return []
	if type(i)==types.ListType:
	    return i
	else:
	    return [i]

#End class RetrieveQuery


#############################################################################
#  DeleteMultipleQuery
#  
#  Query for the purpose of multi-row deletes
#############################################################################
class DeleteMultipleQuery(Query):

    #########################################################################
    # constructor
    #
    #  access     = name or DBAccess object to reference for database access
    #  obj        = object type to delete with this query
    #  criteria   = criteria to use with this query
    #  accessName (deprecated) name to reference for database access
    #########################################################################
    def __init__(self,access,obj=None,criteria=None,accessName=''):
	if not access:
	    access=accessName
	Query.__init__(self,access=access,obj=obj)
	self.obj.setAlias(self.obj.getTableName())


    ########################################################################
    #  getSQL
    #
    #  Return query text
    ########################################################################
    def getSQL(self):
	"""return sql for this query"""

	s = "delete from %s \n" % self.obj.getTableName()
	if self.criteria:
	    conditionText=self.criteria.getSQL()
	    if conditionText:
	        s=s + "where %s\n" % conditionText

	return s

#END class DeleteMultipleQuery


#############################################################################
#  IdLookupQuery
#  
#  Special query for lookup of current value for object id.  This query
#  is both unique enough and commonly used, that it warrants its own class
#############################################################################
class IdLookupQuery(RetrieveQuery):

    ######################################################################
    # constructor
    ######################################################################
    def __init__(self,access,obj=None):
	Query.__init__(self,access,obj)

    ######################################################################
    # getSQL
    ######################################################################
    def getSQL(self,startRow=0,endRow=0):
	'''return sql for this query'''

	if self.isPostgres():
	    s = "select %s" % (self.obj.getIdLookup())
	else:
	    s = "select %s from %s where rownum=1\n" % \
	        (self.obj.getIdLookup(), self.obj.getRootTable())

	return s

#End class IdLookupQuery


#############################################################################
#  SubQuery
#  
#  The subquery class is specifically for queries used within conditions
#  when using the IN or EXISTS operators.  Sub queries differ from outer
#  queries in that they can reference outer query object attributes
#############################################################################
class SubQuery(RetrieveQuery):
    
    #########################################################################
    # constructor
    #
    #  access = name to reference for database access
    #  obj        = object type to create with this query
    #  criteria   = criteria to use with this query
    #  selectColumns = columns to use for selection, default is all
    #########################################################################
    def __init__(self,access='',obj=None,criteria=None,selectColumns=None,
	outerObj=None):
	RetrieveQuery.__init__(self,access,obj,criteria,selectColumns)
	self.outerObj=outerObj
	

    #########################################################################
    # setOuterObject
    #########################################################################
    def setOuterObject(self,obj):
	'''Set reference to outer object'''
	self.outerObj=obj
	if self.criteria:
	    self.criteria.addObject(obj,hidden=1)


    ######################################################################
    #  addCriteria
    #
    #  add criteria to query and return the added condition
    ######################################################################
    def addCriteria(self,left=None,right=None,op='=',negate=0,
		caseSensitive=0,condition=None):
	"""add criteria to query"""
	rv=Query.addCriteria(self,left,right,op,negate,caseSensitive,condition)
	if self.outerObj:
	    self.criteria.addObject(self.outerObj,hidden=1)
	    rv.addObject(self.outerObj,hidden=1)

	return rv


    ########################################################################
    #  getSQL
    #
    #  Return sql for query, optional start and end row parameters
    #  allow for query by row number
    ########################################################################
    def getSQL(self):
	"""return sql for this query"""

	s='select '
	if not self.selectColumns:
	    s = s + self.getColumns() + "\n"              # select
	else:
	    separator=''
	    for i in self.selectColumns:
		s=s+separator+self.obj.getColumnName(i,withAlias=1)
		separator=', '

	s = s + "from " + self.getFromList() + "\n"   # from

	if self.criteria:
	    conditionString = self.criteria.getSQL()
	    if conditionString:
	        s = s + "where " + conditionString + "\n" # where

	return s

#End class SubQuery


#############################################################################
#  OverrideQuery
#
#  class to represent queries that are forced from simple string rather
#  than constructed
#############################################################################
class OverrideQuery(RetrieveQuery):

    #########################################################################
    # constructor
    #########################################################################
    def __init__(self,access=None,obj=None,query='',accessName=''):
	RetrieveQuery.__init__(self,access=access,obj=obj,accessName=accessName)
	self.override=query


    ########################################################################
    #  setQuery
    #
    #  Set the override query string
    ########################################################################
    def setQuery(self,query):
        '''set the override query string'''
	self.override=query


    ########################################################################
    #  getSQL
    #
    #  Return sql for query
    ########################################################################
    def getSQL(self,startRow=0,endRow=0):
	"""return sql for this query"""
	if self.obj:
	    s=self.override % (self.getColumns())
	else:
	    s=self.override
	return s

#END class SubQuery




#----------------------------------------------------------------------------
#--  Condition Classes
#----------------------------------------------------------------------------

###################################################################
#  class CriteriaException
#
###################################################################
class CriteriaException(exceptions.Exception):
    """Exception for problems with query conditions/criteria"""

    def __init__(self,args=None):
        self.args = args

#END class CriteriaException


##############################################################################
#  Condition
#
#  The Condition class is the parent class for the classes necessary for
#  the construction of database selection condition generation
#
#  Subclasses include:  Criteria, Comparison, AndOrExpression, and 
#			OverrideCondition
#
#  Properties:
#	objects = list of objects potentially referenced by this condition
#       negate  = flag to indicated whether condition should be negated or not
#
#  Operations:
#       getSQL = build the condition string
#
##############################################################################
class Condition:

    # Operators for Comparisons
    OR='or'
    AND='and'
    GT='>'
    LT='<'
    GE='>='
    LE='<='
    NE='<>'
    EQ='='
    CONTAINS='contains'
    STARTS_WITH='starts_with'
    IN='in'
    NOT_IN='not in'


    ######################################################################
    # constructor
    ######################################################################
    def __init__(self,obj=[],negate=0,orFlag=0):
	self.setObjects(obj)
	self.negate=negate
	self.orFlag=orFlag


    ######################################################################
    #  dump
    ######################################################################
    def dump(self,indent=''):
	s=indent+"objects: " 
	for i in self.objects:
	    s=s + str(i) + "  "
	print s
	print indent + " negate: %d" % self.negate
	print indent + "     or: %d" % self.orFlag


    ######################################################################
    # __toQueryObject(obj)
    #
    # return query object for given object
    ######################################################################
    def __toQueryObject(self,obj,alias='',hidden=0):
	qObj=None
	if obj and obj.__class__.__name__=='QueryObject':
	    if alias:
	         obj.setAlias(alias)
	    qObj=obj
	else:
	    qObj=QueryObject(obj,alias)

	if hidden:
	    qObj.setHidden()

	return qObj


    ######################################################################
    # Condition::getFrom
    #
    # return dictionary "tablename alias":"alias" for all objects
    # referenced by this condition, if a dictionary is passed in
    # then new references will be added to it
    ######################################################################
    def getFrom(self,dict={}):
	"""return dictionary of tables with aliases referenced by this
           condition"""

	for obj in self.objects:
	    if not obj.isHidden():
	        k="%s %s" % (obj.getViewName(),obj.getAlias())
	        dict[k]=obj.getAlias()

	return dict


    ######################################################################
    # setOr
    ######################################################################
    def setOr(self):
	self.orFlag=1


    ######################################################################
    # isOr
    ######################################################################
    def isOr(self):
	return self.orFlag


    ######################################################################
    # setNegate
    ######################################################################
    def setNegate(self,negate):
	self.negate=negate


    ######################################################################
    # setObjects
    ######################################################################
    def setObjects(self,objects):
	self.objects=[]
	if objects:
	    if type(objects)==types.ListType:
		for i in objects:
	            self.objects.append(self.__toQueryObject(i))
	    else:
	        self.objects.append(self.__toQueryObject(objects))


    ######################################################################
    # addObject
    ######################################################################
    def addObject(self,obj,alias='',hidden=0):
	if alias:
	    qObj=self.__toQueryObject(obj=obj,alias=alias,hidden=hidden)
	    self.objects.append(qObj)
	else:
	    qObj=self.__toQueryObject(obj=obj,hidden=hidden)
	    self.objects.append(qObj)


    ######################################################################
    # getSQL
    ######################################################################
    def getSQL(self,startRow=0,endRow=0):
	'''this method needs to be overridden in subclasses'''
	return ''

  
    ######################################################################
    # __str__
    ######################################################################
    def __str__(self):
        return self.getSQL()

#END class Condition


##############################################################################
#  Criteria
#
#  The Criteria class is responsible for holding query criteria and building
#  the appropriate condition based on that criteria
#
#  Properties:
#	object       - object of type PersistentObject to work with
#       conditions   - list of sub-conditions (comparisons or sub-criteria)
#
#  Operations:
#	add          - add criteria: comparison, or, and
#       getSQL       - return condition text based on critia
#
##############################################################################
class Criteria(Condition):

    ######################################################################
    # constructor
    ######################################################################
    def __init__(self,obj=[],negate=0,conditions=[]):
	Condition.__init__(self,obj=obj,negate=negate)
	if type(conditions)==types.ListType:
	    self.conditions=conditions
	else:
	    self.conditions=[conditions]


    ######################################################################
    # dump
    ######################################################################
    def dump(self,indent=''):
	print indent+"----- Criteria ------"
	Condition.dump(self,indent)
	for i in self.conditions:
	    i.dump('    '+indent)


    ######################################################################
    #  add
    #
    #  add condition to criteria and return the added condition
    ######################################################################
    def add(self,left=None,right=None,op=Condition.EQ,negate=0,
	    caseSensitive=0, condition=None):
	"""add comparison or and/or clause to criteria"""

	rv=None

	if condition:
	    if type(condition)==types.StringType:
	        rv=OverrideCondition(condition)
	    else:
	        rv=condition

	else:
            rv=Comparison(obj=self.objects,left=left,right=right,
		    op=op,negate=negate,caseSensitive=caseSensitive)
	self.conditions.append(rv)

	return rv
	

    ######################################################################
    #  addSearch
    ######################################################################
    def addSearch(self,exact=0,**parms):
	self.addSearchFromDict(exact=exact,parms=parms)
	

    ######################################################################
    #  addSearch
    #
    #  add search criteria,  this method is primarily to provide a
    #  simple interface to add search comparisons.  Only true
    #  parameters will be included in the criteria
    ######################################################################
    def addSearchFromDict(self,exact=0,parms={}):
	for k in parms.keys():
	    value=parms[k]
	    if value:
		condition=Comparison(obj=self.objects,left=k,right=value,
		     caseSensitive=1,exact=exact)
		self.conditions.append(condition)


    ######################################################################
    # addOr
    #
    # adds condition as an 'or' rather than 'and'
    ######################################################################
    def addOr(left=None,right=None,op=Condition.EQ,negate=0,caseSensitive=0,
		condition=None):
	"""add conditon as an 'or' rather than 'and'"""
	rv=self.add(left,right,op,negate,caseSensitive,condition)
	rv.setOr()
	return rv


    ######################################################################
    # Criteria::getSQL
    ######################################################################
    def getSQL(self):
	"""return criteria as string"""
	if not self.conditions:
	    return ""

	if self.negate:
	    s=" not ("
	else:
	    s=""
	prefix=""
	for c in self.conditions:
	    s = s + prefix + "(" + c.getSQL() + ")"
	    prefix=" and "
	if self.negate:
	    s=s+")"
	return s


    ######################################################################
    # Criteria::getFrom
    #
    # return dictionary of "tablename alias":"alias" for all
    # objects referenced by this comparison
    ######################################################################
    def getFrom(self,dict={}):
	"""return dictionary of tables with aliases for all objects
	   referenced by this criteria"""

	Condition.getFrom(self,dict)
	for i in self.conditions:
	    i.getFrom(dict)

	return dict

# END Criteria Class


##############################################################################
#  Comparison
#
#  The Comparison class is responsible for holding a single comparison
#  using the basic comparison operations plus a couple of additional
#  common operations such as contains and starts with
#
#  Properties:
#       left, right: - left and right parts of comparison
#       op           - operator
#
#  Operations:
#       getSQL  - return condition text based on critia
#
##############################################################################
class Comparison(Condition):

    ######################################################################
    # constructor
    ######################################################################
    def __init__(self,obj=None,left=None,right=None,op=Condition.EQ,negate=0, 
		caseSensitive=0, leftForceString=0, rightForceString=0,exact=1):

	# aquired attributes
	self.left=left
	self.right=right
	self.operation=op
	self.caseSensitive=caseSensitive
	self.leftForceString=leftForceString
	self.rightForceString=rightForceString
	Condition.__init__(self,obj=obj,negate=negate)

	# derived attributes
	self.__resolveColumnNames()

	# make adjustments if exact flag turned off
	if exact==0:
	    self.caseSensitive=0
	    if self.compareType == "string" and op==Condition.EQ:
	         self.operation=Condition.STARTS_WITH


    ######################################################################
    # __resolveColumnNames
    #
    # pre:  left and right properties must be set
    # post: leftColumnName, rightColumnName, and compareType will
    #       be adjusted
    #
    # this method should be called any time the object references or 
    # left and right values are changed
    ######################################################################
    def __resolveColumnNames(self):
	self.leftColumnName=''
	self.rightColumnName=''
	if not self.leftForceString and type(self.left)==types.StringType:
	    self.leftColumnName=self.__getColumnName(self.left)
	if not self.rightForceString and type(self.right)==types.StringType:
	    self.rightColumnName=self.__getColumnName(self.right)
	self.compareType=self.__determineCompareType()


    ######################################################################
    # dump
    ######################################################################
    def dump(self,indent=''):
	print indent+"----- Comparison ------"
	Condition.dump(self,indent)

	# left side
	if type(self.left)==types.InstanceType:
	    print indent + "-- left --"
	    if isinstance(self.left,SubQuery):
		print indent+'    '+self.left.getSQL()
	    else:
	        self.left.dump('    '+indent)
	else:
	    print indent + "   left: " + str(self.left)

	# right side
	if type(self.right)==types.InstanceType:
	    print indent + "-- right --"
	    if isinstance(self.right,SubQuery):
		print indent+'    '+self.right.getSQL()
	    else:
	        self.right.dump('    '+indent)
	else:
	    print indent + "  right: " + str(self.right)

	print indent +     "     op: " + self.__getOperationAsString()
	print indent +     "cp type: " + self.compareType


    ######################################################################
    # Comparison::__determineCompareType
    #
    # comparisons have different types, if type is date or string
    # then special processing will be done on comparison
    #
    # return values:
    #    'date'   - if either side is a date attribute
    #    'string' - if both sides are strings or string columns
    #    'int'    - if either side is integer
    #    'float'  - if either side is float
    ######################################################################
    def __determineCompareType(self):
	ltype=''
	rtype=''
	if type(self.left)==types.StringType:
	    if not self.leftForceString:
	        ltype=self.__getColumnType(self.left)
	    if not ltype:
		ltype='string'

	if type(self.right)==types.StringType:
	    if not self.rightForceString:
	        rtype=self.__getColumnType(self.right)
	    if not rtype:
		rtype='string'

	if ltype=='date' or rtype=='date':
	    return 'date'
	elif ltype=='string' and rtype=='string':
	    return 'string'
	elif ltype=='float' or rtype=='float':
	    return 'float'
	elif ltype=='int' or rtype=='int':
	    return 'int'
	elif ltype=='string' or rtype=='string':
	    return 'string'
	else:
	    return ''


    ######################################################################
    # Comparison::getSQL
    ######################################################################
    def getSQL(self):
	"""return comparison as string"""

	left=self.__getOperandAsString(operand=self.left,
		columnName=self.leftColumnName)
	right=self.__getOperandAsString(operand=self.right,
		columnName=self.rightColumnName)
	operation=self.__getOperationAsString()

	if self.negate:
	    s="not(%s %s %s)" % (left,operation,right)
	else:
	    s="%s %s %s" % (left,operation,right)

	return s


    ######################################################################
    # addObject
    ######################################################################
    def addObject(self,obj,alias='',hidden=0):
	Condition.addObject(self,obj,alias,hidden)
	self.__resolveColumnNames()


    ######################################################################
    # setObjects
    ######################################################################
    def setObjects(self,objects):
	Condition.setObjects(self,objects)
	self.__resolveColumnNames()


    ######################################################################
    # Comparison::__getOperandAsString
    #
    #  Returns the operand (no matter what type it is) as a string
    #  Adds upper or date conversions if necessary
    ######################################################################
    def __getOperandAsString(self,operand,columnName):
	s=''

	if columnName:
	    if self.compareType=='string' and not self.caseSensitive:
		s=Query.UPPER % columnName
	    else:
		s=columnName
	elif self.compareType=='string':
	    s=self.__convertString(operand)
	    if self.compareType=='string':
		if self.operation==Condition.CONTAINS:
		    s=Query.LIKE_CONTAINS % (s)
		elif self.operation==Condition.STARTS_WITH:
		    s=Query.LIKE_STARTS_WITH % (s)
	    s="'%s'" % (s)
	    if not self.caseSensitive:
		s=Query.UPPER % (s)
	elif self.compareType=='date':
	    s=Query.STRING_TO_DATE % operand
	elif type(operand)==types.InstanceType and \
	    isinstance(operand,SubQuery):
	    s="(" + operand.getSQL() + ")"
	else:
	    s=str(operand)
		
	return s


    ######################################################################
    # Comparison::__getOperationAsString
    ######################################################################
    def __getOperationAsString(self):
        op=self.operation
	if op==Condition.GE:
	    s='>='
	elif op==Condition.GT:
	    s='>'
	elif op==Condition.LE:
	    s='<='
	elif op==Condition.NE:
	    s='<>'
	elif op==Condition.EQ:
	    s='='
	elif op==Condition.LT:
	    s='<'
	elif op==Condition.IN:
	    s='in'
	elif op==Condition.NOT_IN:
	    s='not in'
	elif op==Condition.CONTAINS:
	    s='like'
	elif op==Condition.STARTS_WITH:
	    s='like'
	else:
	    s=''

	return s



    ######################################################################
    # Comparison::getFrom
    #
    # return dictionary of "tablename alias":"alias" for all
    # objects referenced by this comparison
    ######################################################################
    def getFrom(self,dict={}):
	"""return dictionary of tables with aliases for all objects
	   referenced by this comparison"""

	Condition.getFrom(self,dict)
	#if type(self.left)==types.InstanceType:
	#    self.left.getFrom(dict)

        #if type(self.right)==types.InstanceType:
	#    self.right.getFrom(dict)

	return dict
	

    ######################################################################
    # Comparison::__getColumnName
    ######################################################################
    def __getColumnName(self,name):
	column=''
	if string.find(name,".")>0:
	    parts=string.split(name,".")
	    classname=parts[0]
	    propertyname=parts[1]
	else:
	    classname=''
	    propertyname=name

	# actual column names
	for i in self.objects:
	    if not classname or classname==i.getClassName():
	        column=i.getColumnName(propertyname,withAlias=1)
	    if column:
		return column

	# special purpose column names
	if string.lower(propertyname)=='rownum':
	    column='rownum'

	return column
	

    ######################################################################
    # Comparison::__getColumnType
    ######################################################################
    def __getColumnType(self,name):
	columnType=''
	if string.find(name,".")>0:
	    parts=string.split(name,".")
	    classname=parts[0]
	    propertyname=parts[1]
	else:
	    classname=''
	    propertyname=name

	for i in self.objects:
	    columnType=i.getType(propertyname)
	    if columnType:
		return columnType

	# special purpose column names
	if string.lower(propertyname)=='rownum':
	    columnType='int'

	return columnType
	

    ######################################################################
    #  Comparison::__convertStringForUpdate
    #
    #  The following conversions will be made:
    #      - single quote will be converted to 3 single quotes
    ######################################################################
    def __convertString(self,str):
        s1 = string.replace(str,"'","''")
        return s1


# END Comparison Class


##############################################################################
#  ExistsCondition
#
#  Nested query exists/not exists condition
##############################################################################
class ExistsCondition(Condition):

    ######################################################################
    # constructor
    ######################################################################
    def __init__(self,obj=[],negate=0,orFlag=0,query=None):
	Condition.__init__(self,obj=obj,negate=negate,orFlag=orFlag)
	for i in self.objects:
	    i.setHidden()
	self.query=query


    ######################################################################
    # getSQL
    ######################################################################
    def getSQL(self):
	"""return SQL for this condition"""
	if self.negate:
	    s="not exists ( %s )" % self.query.getSQL()
	else:
	    s="exists ( %s )" % self.query.getSQL()

	return s

 
    ######################################################################
    # setQuery
    ######################################################################
    def setQuery(self,query):
  	'''Set the subquery for this exists condition'''
	self.query=query


    ######################################################################
    # dump
    ######################################################################
    def dump(self,indent=''):
	print indent+"----- Exists Condition -----"
	print indent+"query: " + self.query.getSQL()

#END class ExistsCondition


##############################################################################
#  OverrideCondition
#
#  Condition composed of string
##############################################################################
class OverrideCondition(Condition):

    ######################################################################
    # constructor
    ######################################################################
    def __init__(self,obj=[],condition='',negate=0,orFlag=0):
	Condition.__init__(self,obj=obj,negate=negate,orFlag=orFlag)
	self.override=condition


    ######################################################################
    # getSQL
    ######################################################################
    def getSQL(self):
	"""return SQL for this condition"""
	if self.negate:
	    s="not( %s )" % self.override
	else:
	    s=self.override

	return s

    ######################################################################
    # dump
    ######################################################################
    def dump(self,indent=''):
	print indent+"----- Override Condition ------"
	print indent+"override: " + self.override

#END class OverrideCondition

